// -*- c++ -*-
/*
 *  MICO --- an Open Source CORBA implementation
 *  Copyright (c) 1997-2005 by The Mico Team
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  For more information, visit the MICO Home Page at
 *  http://www.mico.org/
 */

#ifndef __mico_address_impl_h__
#define __mico_address_impl_h__

namespace MICO {

class MICO_EXPORT LocalAddress : public CORBA::Address {
public:
    LocalAddress ();
    ~LocalAddress ();

    std::string stringify () const;
    const char *proto () const;
    CORBA::Transport *make_transport () const;
    CORBA::TransportServer *make_transport_server () const;
    CORBA::IORProfile *make_ior_profile (
	CORBA::Octet *, CORBA::ULong, const CORBA::MultiComponent &,
        CORBA::UShort version = 0x0100) const;
    CORBA::Boolean is_local () const;
    CORBA::Boolean is_here () const;

    CORBA::Address *clone () const;
    
    CORBA::Long compare (const CORBA::Address &) const;
    CORBA::Boolean operator== (const CORBA::Address &) const;
    CORBA::Boolean operator< (const CORBA::Address &) const;
};

class MICO_EXPORT LocalAddressParser : public CORBA::AddressParser {
public:
    LocalAddressParser ();
    ~LocalAddressParser ();
    CORBA::Address *parse (const char *, const char *) const;
    CORBA::Boolean has_proto (const char *) const;
};


class InetAddress : public CORBA::Address {
public:
    enum Family {
	STREAM,
	DGRAM
    };
private:
    CORBA::UShort _port;
    std::string _host;
    std::vector<CORBA::Octet> _ipaddr;
    Family _family;
    static CORBA::Boolean _resolve;

    static std::string hname;
    static std::vector<CORBA::Octet> hid;

    MICO_EXPORT CORBA::Boolean resolve_ip () const;
    MICO_EXPORT CORBA::Boolean resolve_host () const;
    // kcg: temporary lock for gethostbyname/gethostbyaddr
    // calls serialization. InetAddress and Transports
    // should be rewritten to use getaddrinfo/getnameinfo
    // functions
    static MICOMT::Mutex S_netdb_lock_;
public:
    MICO_EXPORT InetAddress (struct sockaddr_in &sin, Family = STREAM);
    MICO_EXPORT InetAddress (const char *host = NULL, CORBA::UShort port = 0,
		 Family = STREAM);
    MICO_EXPORT InetAddress (const std::vector<CORBA::Octet> &, CORBA::UShort port = 0,
		 Family = STREAM);
    MICO_EXPORT ~InetAddress ();

    MICO_EXPORT std::string stringify () const;
    MICO_EXPORT const char *proto () const;
    MICO_EXPORT CORBA::Transport *make_transport () const;
    MICO_EXPORT CORBA::TransportServer *make_transport_server () const;
    MICO_EXPORT CORBA::IORProfile *make_ior_profile (
	CORBA::Octet *, CORBA::ULong, const CORBA::MultiComponent &,
        CORBA::UShort version = 0x0100) const;
    MICO_EXPORT CORBA::Boolean is_local () const;
    MICO_EXPORT CORBA::Boolean is_here () const;
    
    MICO_EXPORT CORBA::Address *clone () const;
    
    MICO_EXPORT CORBA::Long compare (const CORBA::Address &) const;
    MICO_EXPORT CORBA::Boolean operator== (const CORBA::Address &) const;
    MICO_EXPORT CORBA::Boolean operator< (const CORBA::Address &) const;

    MICO_EXPORT CORBA::Boolean valid () const;

    MICO_EXPORT const char *host () const;
    MICO_EXPORT void host (const char *);

    MICO_EXPORT const std::vector<CORBA::Octet> &ipaddr () const;
    MICO_EXPORT void ipaddr (const std::vector<CORBA::Octet> &);

    MICO_EXPORT CORBA::UShort port () const;
    MICO_EXPORT void port (CORBA::UShort port);

    MICO_EXPORT void sockaddr (struct sockaddr_in &);
    MICO_EXPORT struct sockaddr_in sockaddr () const;

    MICO_EXPORT void family (Family f);
    MICO_EXPORT Family family () const;

    MICO_EXPORT static std::string hostname ();
    MICO_EXPORT static std::vector<CORBA::Octet> hostid();
    MICO_EXPORT static CORBA::Boolean samehosts (const std::string &h1,
				     const std::string &h2);

    MICO_EXPORT static CORBA::Boolean resolve ();
    MICO_EXPORT static void resolve (CORBA::Boolean);
};

class InetAddressParser : public CORBA::AddressParser {
public:
    MICO_EXPORT InetAddressParser ();
    MICO_EXPORT ~InetAddressParser ();
    MICO_EXPORT CORBA::Address *parse (const char *, const char *) const;
    MICO_EXPORT CORBA::Boolean has_proto (const char *) const;
};

#ifdef HAVE_SYS_UN_H
class UnixAddress : public CORBA::Address {
    std::string _filename;
public:
    MICO_EXPORT UnixAddress (struct sockaddr_un &una);
    MICO_EXPORT UnixAddress (const char *filename = 0);
    MICO_EXPORT ~UnixAddress ();

    MICO_EXPORT std::string stringify () const;
    MICO_EXPORT const char *proto () const;
    MICO_EXPORT CORBA::Transport *make_transport () const;
    MICO_EXPORT CORBA::TransportServer *make_transport_server () const;
    MICO_EXPORT CORBA::IORProfile *make_ior_profile (
	CORBA::Octet *, CORBA::ULong, const CORBA::MultiComponent &,
        CORBA::UShort version = 0x0100) const;
    MICO_EXPORT CORBA::Boolean is_local () const;
    MICO_EXPORT CORBA::Boolean is_here () const;
    
    MICO_EXPORT CORBA::Address *clone () const;
    
    MICO_EXPORT CORBA::Long compare (const CORBA::Address &) const;
    MICO_EXPORT CORBA::Boolean operator== (const CORBA::Address &) const;
    MICO_EXPORT CORBA::Boolean operator< (const CORBA::Address &) const;

    MICO_EXPORT const char *filename () const;
    MICO_EXPORT void filename (const char *);

    MICO_EXPORT void sockaddr (struct sockaddr_un &);
    MICO_EXPORT struct sockaddr_un sockaddr () const;
};

class UnixAddressParser : public CORBA::AddressParser {
public:
    MICO_EXPORT UnixAddressParser ();
    MICO_EXPORT ~UnixAddressParser ();
    MICO_EXPORT CORBA::Address *parse (const char *, const char *) const;
    MICO_EXPORT CORBA::Boolean has_proto (const char *) const;
};
#endif // HAVE_SYS_UN_H

}

#endif // __mico_address_impl_h__
