// -*- c++ -*-
/*
 *  MICO --- an Open Source CORBA implementation
 *  Copyright (c) 1997-2001 by The Mico Team
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  For more information, visit the MICO Home Page at
 *  http://www.mico.org/
 */

#ifndef __mico_typecode_h__
#define __mico_typecode_h__

#include <mico/export.h>

namespace CORBA {

class TypeCode : public ServerlessObject {
    TCKind tckind;
    std::string repoid;
    std::string tcname;
    std::vector<std::string> namevec;
    std::vector<TypeCode_ptr> tcvec;
    std::vector<Any *> labelvec;
    std::vector<Visibility> visvec;
    TypeCode_ptr content;
    TypeCode_ptr discriminator;
    TypeCode_ptr recurse_tc;
    Long defidx;
    Long len;
    Long recurse_depth;
    Boolean isconst;
    UShort digits;
    Short scale;
    ValueModifier value_mod;

    MICO_EXPORT void init ();
    MICO_EXPORT void copy (const TypeCode &);
    MICO_EXPORT void free ();

    typedef std::pair<TypeCode_ptr, TypeCode_ptr> PairTC;
    typedef std::set<PairTC, std::less<PairTC> > SetTC;
    typedef std::map<ULong, std::pair<ULong, TypeCode_ptr>, std::less<ULong> >
    MapPosTC;
    typedef std::map<const TypeCode *, ULong, std::less<const TypeCode *> >
    MapTCPos;

    MICO_EXPORT TypeCode_ptr resolve_recursion() const;
    MICO_EXPORT void strip();
    MICO_EXPORT void connect (TypeCode_ptr, Long depth = 0);
    MICO_EXPORT CORBA::ULong get_recurse_depth();
    MICO_EXPORT static CORBA::Any *convert_case_label (CORBA::TypeCode_ptr disc,
					   const CORBA::Any &);
public:
    // begin-mico-extension
    MICO_EXPORT void disconnect (TypeCode_ptr);
    MICO_EXPORT TypeCode (const std::string &);
    MICO_EXPORT TypeCode (); // for vector<TypeCode>
    MICO_EXPORT ~TypeCode ();
    MICO_EXPORT TypeCode (TCKind);
    MICO_EXPORT TypeCode (const TypeCode &);
    MICO_EXPORT TypeCode (DataDecoder &);
    MICO_EXPORT TypeCode &operator= (const TypeCode &);
    MICO_EXPORT Boolean from_string (const char *);

    MICO_EXPORT static TypeCode_ptr create_basic_tc (TCKind);
    MICO_EXPORT static TypeCode_ptr create_struct_tc (const char *rep_id, const char *name,
					  const StructMemberSeq &members);
    MICO_EXPORT static TypeCode_ptr create_exception_tc (const char *rep_id,
					     const char *name,
					     const StructMemberSeq &members);
    MICO_EXPORT static TypeCode_ptr create_union_tc (const char *rep_id, const char *name,
					 TypeCode_ptr disc_type,
					 const UnionMemberSeq &members);
    MICO_EXPORT static TypeCode_ptr create_enum_tc (const char *rep_id, const char *name,
					const EnumMemberSeq &members);
    MICO_EXPORT static TypeCode_ptr create_alias_tc (const char *rep_id, const char *name,
					 TypeCode_ptr orig_type);
    MICO_EXPORT static TypeCode_ptr create_interface_tc (const char *rep_id,
					     const char *name);
    MICO_EXPORT static TypeCode_ptr create_string_tc (ULong bound);
    MICO_EXPORT static TypeCode_ptr create_wstring_tc (ULong bound);
    MICO_EXPORT static TypeCode_ptr create_sequence_tc (ULong bound, TypeCode_ptr el_type);
    MICO_EXPORT static TypeCode_ptr create_array_tc (ULong bound, TypeCode_ptr el_type);
    MICO_EXPORT static TypeCode_ptr create_recursive_sequence_tc (ULong bound,
						      ULong offset);
    MICO_EXPORT static TypeCode_ptr create_fixed_tc (UShort digits, Short scale);
    MICO_EXPORT static TypeCode_ptr create_value_tc (const char *rep_id, const char *name,
					 ValueModifier mod,
					 TypeCode_ptr concrete_base,
					 const ValueMemberSeq &members);
    MICO_EXPORT static TypeCode_ptr create_value_box_tc (const char *rep_id,
					     const char *name,
					     TypeCode_ptr boxed_type);
    MICO_EXPORT static TypeCode_ptr create_native_tc (const char *rep_id,
					  const char *name);
    MICO_EXPORT static TypeCode_ptr create_recursive_tc (const char *rep_id);
    MICO_EXPORT static TypeCode_ptr create_abstract_interface_tc (const char *rep_id,
						      const char *name);
    MICO_EXPORT static TypeCode_ptr create_local_interface_tc (const char *rep_id,
						   const char *name);
    // end-mico-extension

    typedef CORBA::TypeCode_Bounds Bounds;
    typedef CORBA::TypeCode_Bounds_catch Bounds_catch;

    typedef CORBA::TypeCode_BadKind BadKind;
    typedef CORBA::TypeCode_BadKind_catch BadKind_catch;

    MICO_EXPORT TCKind kind () const;
    MICO_EXPORT Boolean equal (TypeCode_ptr, Boolean remove_aliases = FALSE,
		                 Boolean ignore_string_bounds = FALSE) const;
    MICO_EXPORT Boolean equaltype (TypeCode_ptr, SetTC * = 0);
    MICO_EXPORT Boolean equivalent (TypeCode_ptr tc);

    MICO_EXPORT const char *id () const;
    MICO_EXPORT const char *name () const;

    MICO_EXPORT Long member_index (const Any &);
    MICO_EXPORT Long member_index (const char *);
    MICO_EXPORT ULong member_count () const;
    MICO_EXPORT const char *member_name (ULong idx) const;
    MICO_EXPORT TypeCode_ptr member_type (ULong idx) const;
    MICO_EXPORT Any *member_label (ULong idx) const;
    MICO_EXPORT TypeCode_ptr discriminator_type () const;
    MICO_EXPORT Long default_index () const;

    MICO_EXPORT ULong length () const;
    
    MICO_EXPORT TypeCode_ptr content_type () const;

    MICO_EXPORT UShort fixed_digits () const;
    MICO_EXPORT Short fixed_scale () const;

    MICO_EXPORT Visibility member_visibility (ULong idx) const;
    MICO_EXPORT ValueModifier type_modifier() const;
    MICO_EXPORT TypeCode_ptr concrete_base_type() const;
    MICO_EXPORT TypeCode_ptr get_compact_typecode() const;

    MICO_EXPORT ULong member_count_inherited () const;
    MICO_EXPORT const char *member_name_inherited (ULong idx) const;
    MICO_EXPORT TypeCode_ptr member_type_inherited (ULong idx) const;
    MICO_EXPORT Visibility member_visibility_inherited (ULong idx) const;

    MICO_EXPORT Long param_count () const;
    MICO_EXPORT Any *parameter (Long) const;

    // begin-mico-extension
    MICO_EXPORT ULong array_size ();
    MICO_EXPORT TypeCode_ptr array_type ();
    MICO_EXPORT Boolean is_constant () const;
    MICO_EXPORT TypeCode_ptr mk_constant ();
    MICO_EXPORT TypeCode_ptr unalias ();

    MICO_EXPORT void encode (DataEncoder &, MapTCPos * = NULL) const;
    MICO_EXPORT Boolean decode (DataDecoder &, MapPosTC * = NULL, ULong level = 0);

    MICO_EXPORT std::string stringify () const;

    MICO_EXPORT Boolean is_recursive_seq ();
    MICO_EXPORT Boolean is_freestanding();
    MICO_EXPORT Boolean is_freestanding( ULong depth, TypeCode_ptr tc );


    MICO_EXPORT Boolean is_base_type();
    MICO_EXPORT Boolean is_char();
    MICO_EXPORT Boolean is_wchar();
    MICO_EXPORT Boolean is_octet();
    MICO_EXPORT Boolean is_bool();
    MICO_EXPORT Boolean is_enum();
    MICO_EXPORT Boolean is_string();
    MICO_EXPORT Boolean is_wstring();
    MICO_EXPORT Boolean is_any();
    MICO_EXPORT Boolean is_type_code();
    MICO_EXPORT Boolean is_alias();
    MICO_EXPORT Boolean is_variable();
    MICO_EXPORT Boolean is_objref();
    MICO_EXPORT Boolean is_object();
    MICO_EXPORT Boolean is_sequence();
    MICO_EXPORT Boolean is_array();
    MICO_EXPORT Boolean is_valuetype();
    MICO_EXPORT Boolean is_native();
    MICO_EXPORT Boolean is_abstract_interface();
    // end-mico-extension

    MICO_EXPORT static TypeCode_ptr _duplicate (TypeCode_ptr);
    static TypeCode_ptr _nil ()
    {
	return 0;
    }
};
MICO_EXPORT void release (TypeCode_ptr);

//typedef ObjVar<TypeCode> TypeCode_var;


// begin-mico-extension
class LevelRecord {
public:
    enum Level {
	LNone, LArray, LSequence, LStruct, LExcept, LUnion, LValue, LValueBox
    };
private:
    TypeCode_ptr _tc;
    Level _level;
public:
    Long i, n, x, last_i;

    MICO_EXPORT LevelRecord (Level, TypeCode_ptr, Long n = 0, Long i = 0, Long x = 0);
    MICO_EXPORT LevelRecord (const LevelRecord &);
    MICO_EXPORT LevelRecord ();
    MICO_EXPORT ~LevelRecord ();
    MICO_EXPORT LevelRecord &operator= (const LevelRecord &);
    MICO_EXPORT Level level () const;
    MICO_EXPORT TypeCode_ptr tc ();
};

class TypeCodeChecker {
    TypeCode_ptr _top;
    TypeCode_ptr _tc;
    Boolean done;
    std::vector<LevelRecord> levelvec;

    MICO_EXPORT LevelRecord *level ();
    MICO_EXPORT Boolean advance ();
    MICO_EXPORT Boolean nexttc ();
    MICO_EXPORT Boolean leave (LevelRecord::Level, Boolean sloppy = FALSE);
    MICO_EXPORT void enter (const LevelRecord &);
    MICO_EXPORT void tc (TypeCode_ptr);
    MICO_EXPORT void override_tc (TypeCode_ptr);
public:
    MICO_EXPORT TypeCodeChecker ();
    MICO_EXPORT TypeCodeChecker (TypeCode_ptr);
    MICO_EXPORT TypeCodeChecker (const TypeCodeChecker &);
    MICO_EXPORT TypeCodeChecker &operator= (const TypeCodeChecker &);
    MICO_EXPORT ~TypeCodeChecker ();

    MICO_EXPORT TypeCode_ptr tc ();
    MICO_EXPORT Boolean inside (TCKind);
    MICO_EXPORT Boolean level_finished ();

    MICO_EXPORT Boolean basic (TypeCode_ptr);
    MICO_EXPORT Boolean enumeration (Long val);
    MICO_EXPORT Boolean abstract_interface (TypeCode_ptr override);

    MICO_EXPORT Boolean union_begin ();
    MICO_EXPORT Boolean union_selection (Long idx = -1L);
    MICO_EXPORT Boolean union_end (Boolean sloppy = FALSE);

    MICO_EXPORT Boolean struct_begin ();
    MICO_EXPORT Boolean struct_end (Boolean sloppy = FALSE);

    MICO_EXPORT Boolean except_begin ();
    MICO_EXPORT Boolean except_end (Boolean sloppy = FALSE);

    MICO_EXPORT Boolean seq_begin (ULong len);
    MICO_EXPORT Boolean seq_end (Boolean sloppy = FALSE);

    MICO_EXPORT Boolean arr_begin ();
    MICO_EXPORT Boolean arr_end (Boolean sloppy = FALSE);

    MICO_EXPORT Boolean value_begin (TypeCode_ptr override = 0);
    MICO_EXPORT Boolean value_end ();

    MICO_EXPORT Boolean valuebox_begin ();
    MICO_EXPORT Boolean valuebox_end ();

    MICO_EXPORT void finish ();
    MICO_EXPORT Boolean completed () const;
    MICO_EXPORT void restart (TypeCode_ptr);
    MICO_EXPORT void restart ();
    MICO_EXPORT ULong level_count () const;
};
// end-mico-extension


extern MICO_EXPORT TypeCode_ptr _tc_null;
extern MICO_EXPORT TypeCode_ptr _tc_void;
extern MICO_EXPORT TypeCode_ptr _tc_short;
extern MICO_EXPORT TypeCode_ptr _tc_long;
extern MICO_EXPORT TypeCode_ptr _tc_longlong;
extern MICO_EXPORT TypeCode_ptr _tc_ushort;
extern MICO_EXPORT TypeCode_ptr _tc_ulong;
extern MICO_EXPORT TypeCode_ptr _tc_ulonglong;
extern MICO_EXPORT TypeCode_ptr _tc_float;
extern MICO_EXPORT TypeCode_ptr _tc_double;
extern MICO_EXPORT TypeCode_ptr _tc_longdouble;
extern MICO_EXPORT TypeCode_ptr _tc_boolean;
extern MICO_EXPORT TypeCode_ptr _tc_char;
extern MICO_EXPORT TypeCode_ptr _tc_wchar;
extern MICO_EXPORT TypeCode_ptr _tc_octet;
extern MICO_EXPORT TypeCode_ptr _tc_any;
extern MICO_EXPORT TypeCode_ptr _tc_TypeCode;
extern MICO_EXPORT TypeCode_ptr _tc_Principal;
extern MICO_EXPORT TypeCode_ptr _tc_Context; // extension
extern MICO_EXPORT TypeCode_ptr _tc_NamedValue; // extension
extern MICO_EXPORT TypeCode_ptr _tc_Object;
extern MICO_EXPORT TypeCode_ptr _tc_ValueBase;
extern MICO_EXPORT TypeCode_ptr _tc_AbstractBase; // extension
extern MICO_EXPORT TypeCode_ptr _tc_string;
extern MICO_EXPORT TypeCode_ptr _tc_wstring;

}

#endif // __mico_typecode_h__
